import pytest

from tackle.models import Context
from tackle.render import render_string


RENDERABLES = [
    # This test relates to https://github.com/pallets/jinja/issues/1580 where if we have
    # a variable that collides with a jinja globals it is ignored
    # https://github.com/robcxyz/tackle-box/issues/19
    ({'dict': {'stuff': 'things'}}, '{{dict}}', {'stuff': 'things'}),
    ({'namespace': {'stuff': 'things'}}, '{{namespace}}', {'stuff': 'things'}),
    # Normal
    ({'adict': {'stuff': 'things'}}, '{{adict}}', {'stuff': 'things'}),
    ({'list': ['stuff', 'things']}, '{{list}}', ['stuff', 'things']),
]


@pytest.mark.parametrize("context,raw,expected_output", RENDERABLES)
def test_render_string_inputs(context, raw, expected_output):
    context = Context(public_context=context)
    output = render_string(context=context, raw=raw)

    assert output == expected_output



"""Meta hooks."""
import pathlib
import os
import re
import subprocess
from collections import MutableMapping
from PyInquirer import prompt
from pydantic import BaseModel, ValidationError, validator
from typing import Optional

from tackle.utils.paths import work_in
from tackle.models import BaseHook, Field


class Repo(BaseModel):
    """Repo class for parsing dict inputs for repo."""

    src: str
    branch: Optional[str] = None
    tag: Optional[str] = None


# https://stackoverflow.com/a/6027615/12642712
def flatten_repo_tree(d, parent_key=''):
    """Flatten a dict to so that keys become nodes in a path."""
    items = []
    for k, v in d.items():
        new_key = os.path.join(parent_key, k)

        try:
            repo = Repo(**v)
        except (ValidationError, TypeError):
            repo = None

        if repo:
            items.append((new_key, v))
        elif isinstance(v, MutableMapping):
            items.extend(flatten_repo_tree(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


class MetaGitHook(BaseHook):
    """
    Hook to create meta repo, ie a repo that has many other repos within it. See
     [meta repo definition.](https://notes.burke.libbey.me/metarepo/#:~:text=a%20%E2%80%9CMetarepo%E2%80%9D.-,Metarepo%20Architecture,branch%20commit%20for%20each%20repository.)
     Describe the repo structure with a map and when this hook is called, it allows the
     users to operate git commands on that repo structure.
    """

    __slots__ = ('first_run',)
    # Per https://github.com/samuelcolvin/pydantic/issues/655 for private vars

    hook_type: str = 'meta_repo'
    command: str = Field(None, description="The git command to run.")
    repo_tree: dict = Field(
        None,
        description="A map of repos to clone with the keys indicating the path "
        "and values for the repo targets. See examples.",
    )
    protocol: str = Field(
        "https",
        description="Either https or ssh. Useful when using abbreviated repo "
        "references.",
    )
    token: str = Field(
        None, description="A token to use with https abbreviated repo references."
    )
    base_url: str = Field(
        "github.com",
        description="The base url to use with https abbreviated repo " "references.",
    )
    git_org: str = Field(
        None,
        description="The git org to override with https abbreviated repo "
        "references.",
    )

    _doc_tags: list = ["experimental"]
    _issue_numbers: list = []

    @validator('token')
    def no_token_with_ssh_protocol(cls, v, values, **kwargs):
        if v and values['protocol'] == 'ssh':
            raise ValueError("Can't supply token with ssh protocol.")
        else:
            return v

    @validator('base_url')
    def update_base_url_if_token_exists(cls, v, values, **kwargs):
        if values['token']:
            return values['token'] + ':x-oauth-basic@' + v
        else:
            return v

    def get_repo_prefix(self, repo):
        """Return a string to prefix url."""
        if repo.startswith("https") or repo.startswith("git"):
            # Case where it is provided
            return ""
        if self.protocol == "https":
            return f"https://{self.base_url}/"
        if self.protocol == "ssh":
            return f"git@{self.base_url}:"

    def get_git_command(self, branch, repo, folder_name):
        """Build string to run git command."""
        if self.command == 'clone':
            if not branch:
                return f"git {self.command} {self.get_repo_prefix(repo)}{repo} {folder_name}"
            else:
                return (
                    f"git {self.command} {self.get_repo_prefix(repo)}{repo} "
                    f"{folder_name} -b {branch}"
                )
        elif self.command == 'pull':
            return f"git {self.command}"

        # if not branch:
        #     return (
        #         f"git {self.command} {self.get_repo_prefix(repo)}{repo} {folder_name}"
        #     )
        # else:
        #     return (
        #         f"git {self.command} {self.get_repo_prefix(repo)}{repo} "
        #         f"{folder_name} -b {branch}"
        #     )

    def execute_git_command(self, branch, repo, folder_name):
        """Execute the git command."""
        cmd = self.get_git_command(branch, repo, folder_name)
        p = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        output, err = p.communicate()

        if err:
            print(f"{err} for repo='{repo}' in folder='{folder_name}'.")

    def try_https_then_ssh(self, branch, repo, folder_name):
        """When running first git command, tries with https then falls back to ssh."""
        try:
            self.execute_git_command(branch, repo, folder_name)
            object.__setattr__(self, 'first_run', False)
        except Exception:
            try:
                self.protocol = "ssh"
                self.execute_git_command(branch, repo, folder_name)
            except Exception as e:
                print(f"{e} for repo='{repo}' - Neither https or ssh works. Skipping.")

    def run_git_command(self, repo, folder_name, branch=None):
        """Run the git command."""
        if self.first_run:
            self.try_https_then_ssh(branch, repo, folder_name)
        self.execute_git_command(branch, repo, folder_name)

    def update_repo_name(self, v):
        """Update the repo name."""
        if re.compile(r"(https:\/\/[\w\.]+)|(git@[\w\.]+)").match(v):
            return v

        git_parts = v.split('/')
        if len(git_parts) == 1:
            if not self.git_org:
                print(
                    f"No git org declared - only one item declared in {v}, "
                    f"must be in form org/repo for abbreviated declarations. Skipping."
                )
                return
            else:
                return f"{self.git_org}/{git_parts[0]}"
        if len(git_parts) == 2:
            return f"{git_parts[0]}/{git_parts[1]}"
        else:
            from tackle.utils.dicts import get_key_from_key_path

            key = get_key_from_key_path(self.key_path)
            print(f"Malformed repo name '{v}' in '{key}' key. Skipping.")

    def prompt_repo_choices(self):
        """Prompt the user to select which items to operate on."""
        choices = [
            {"name": f"{k} --> {v}", "checked": True} for k, v in self.repo_tree.items()
        ]
        question = {
            'type': 'checkbox',
            'name': 'tmp',
            'message': f"Which repos do you want to {self.command.split(' ')[0]}?",
            'choices': choices,
        }
        repo_choices = prompt([question])['tmp']
        self.repo_tree = {
            k: v for k, v in self.repo_tree.items() if f"{k} --> {v}" in repo_choices
        }

    def get_command(self):
        """Get the git command to run."""
        if not self.no_input:
            git_commands = [
                '<custom input>',
                'clone',
                'pull',
                'branch',
            ]

            question = {
                'type': 'list',
                'name': 'tmp',
                'message': "What to do?",
                'choices': git_commands,
            }
            self.command = prompt([question])['tmp']

            # Extra input
            if self.command == '<custom input>':
                message = "Enter git command: "
                question = {
                    'type': 'input',
                    'name': 'tmp',
                    'message': message,
                }
                self.command = prompt([question])['tmp']
            if self.command == 'branch':
                message = "Enter git branch: "
                question = {
                    'type': 'input',
                    'name': 'tmp',
                    'message': message,
                }
                self.command = "branch " + prompt([question])['tmp']

        else:
            # print(
            #     f"No command given in mete_hook for key = {self.key}.  "
            #     f"Defaults to clone."
            # )
            self.command = 'clone'

    def exec(self):
        """Run the hook."""
        object.__setattr__(self, 'first_run', True)

        if not self.command:
            self.get_command()

        # Flatten the repo tree = `{k1: {k2: v}}` -> `[{k1/k2: v}]` where `k1/k2`
        # is the path to operate (ie clone, etc) on the repo object `v` that can be
        # a string or dict
        if self.repo_tree:
            self.repo_tree = flatten_repo_tree(self.repo_tree)

        # Select is a bool that can override whether to prompt
        if not self.no_input:
            self.prompt_repo_choices()

        # Parse flattened repo tree from above
        for k, v in self.repo_tree.items():
            # Create dir to base path if it does not exist
            path, folder_name = os.path.split(k)
            if not os.path.isdir(path):
                pathlib.Path(path).mkdir(parents=True, exist_ok=True)

            # Handle string inputs
            if isinstance(v, str):
                v = self.update_repo_name(v)
                if path != "":
                    with work_in(path):
                        self.run_git_command(v, folder_name)
                else:
                    self.run_git_command(v, folder_name)

            if isinstance(v, dict):
                r = Repo(**v)
                r.src = self.update_repo_name(r.src)
                if path != "":
                    with work_in(path):
                        self.run_git_command(r.src, folder_name, r.branch)
                else:
                    self.run_git_command(r.src, folder_name, r.branch)



"""Tests `import` in the `tackle.providers.tackle.hooks.match` hook."""
from tackle import tackle


def test_provider_logic_type(change_dir):
    """Check assertions."""
    output = tackle('type.yaml')
    assert output['str_type'] == 'str'
    assert output['list_type'] == 'list'
    assert output['a_dict_type'] == 'dict'
    assert output['dict_list_type'] == 'dict'
    assert output['list_dict_type'] == 'list'



"""Utils."""
import copy
from functools import wraps
import errno
import os
import signal
import ast
import re


def merge_configs(default, overwrite):
    """Recursively update a dict with the key/value pair of another.

    Dict values that are dictionaries themselves will be updated, whilst
    preserving existing keys.
    """
    new_config = copy.deepcopy(default)

    for k, v in overwrite.items():
        # Make sure to preserve existing items in
        # nested dicts, for example `abbreviations`
        if isinstance(v, dict):
            new_config[k] = merge_configs(default[k], v)
        else:
            new_config[k] = v

    return new_config


# https://stackoverflow.com/a/2282656/12642712
class TimeoutError(Exception):
    """Exception for timeouts."""

    pass


def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
    """Timeout for any requests that need decorator."""

    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wraps(func)(wrapper)

    return decorator


def literal_type(input):
    """Take in any serializable argument as string and returns the literal."""
    if isinstance(input, str):
        REGEX = [
            r'^\[.*\]$',  # List
            r'^\{.*\}$',  # Dict
            r'^True$|^False$',  # Boolean
            r'^\d+$',  # Integer
            r'^[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)$',  # Float
        ]
        for r in REGEX:
            # First try to match on a list of regexs that can be evaluated by ast
            if re.match(r, input):
                """If variable looks like list, return literal list."""
                return ast.literal_eval(input)
        return input
    return input



"""All exceptions used in the tackle box code base are defined here."""
import inspect
import os
import sys
from typing import TYPE_CHECKING

from tackle.utils.dicts import get_readable_key_path

if TYPE_CHECKING:
    from typing import Type
    from tackle.models import Context, BaseHook, BaseFunction


#
# Base exceptions - Ones that are subclassed later
#
class ContributionNeededException(Exception):
    """
    Special exception to point users (particularly of the windows variety) to contribute
    a fix with context that includes a link to the file needing upgrade.
    """

    def __init__(self, extra_message=None):
        # self.path_to_code = link_to_code
        self.extra_message = extra_message
        self.message = (
            f"Unimplemented / needs to be built "
            f"- PLEASE HELP -> "
            f"https://github.com/robcxyz/tackle-box/blob/main/{self.file_location()}"
            f"\n{self.extra_message}"
        )
        super().__init__(self.message)

    def __new__(cls, *args, **kwargs):
        new_instance = super(ContributionNeededException, cls).__new__(
            cls, *args, **kwargs
        )
        stack_trace = inspect.stack()
        created_at = '%s:%d' % (stack_trace[1][1], stack_trace[1][2])
        new_instance.created_at = created_at
        return new_instance

    def file_location(self):
        path_list = os.path.normpath(self.created_at).split(os.sep)
        for i, v in enumerate(path_list):
            if v == 'tackle':
                break
        github_path_list = path_list[i:-1] + [
            path_list[-1].split(':')[0] + "#L" + path_list[-1].split(':')[1]
        ]
        return os.path.join(*github_path_list)


# TODO: RM
class TackleException(Exception):
    """
    Base exception class.

    All Tackle-specific exceptions should subclass this class.
    """


class TackleFunctionCallException(Exception):
    """Base hook call exception class."""

    def __init__(self, extra_message: str, function: 'Type[BaseFunction]' = None):
        if function:
            self.message = (
                f"Error parsing input_file='{function.calling_file}' at "
                f"key_path='{get_readable_key_path(key_path=function.key_path)}' \n"
                f"{extra_message}"
            )
            if not function.verbose:
                sys.tracebacklimit = 0
        else:
            self.message = extra_message
        super().__init__(self.message)


class TackleHookCallException(Exception):
    """Base hook call exception class."""

    def __init__(self, extra_message: str, hook: 'Type[BaseHook]' = None):
        if hook:
            self.message = (
                f"Error parsing input_file='{hook.calling_file}' at "
                f"key_path='{get_readable_key_path(key_path=hook.key_path)}' \n"
                f"{extra_message}"
            )
            if not hook.verbose:
                sys.tracebacklimit = 0
        else:
            self.message = extra_message
        super().__init__(self.message)


class TackleParserException(Exception):
    """Base parser exception class."""

    def __init__(self, extra_message: str, context: 'Context' = None):
        if context:
            self.message = (
                f"Error parsing input_file='{context.current_file}' at "
                f"key_path='{get_readable_key_path(key_path=context.key_path)}' \n"
                f"{extra_message}"
            )
            if not context.verbose:
                sys.tracebacklimit = 0
        else:
            self.message = extra_message
        super().__init__(self.message)


class TackleParserInputException(Exception):
    """Base input parser exception class."""

    def __init__(self, extra_message: str, context: 'Context' = None):
        if context:
            self.message = (
                f"Error parsing input_file='{context.input_file}' \n" f"{extra_message}"
            )
            if not context.verbose:
                sys.tracebacklimit = 0
        else:
            self.message = extra_message
        super().__init__(self.message)


#
# Function call exceptions
#
class FunctionCallException(TackleFunctionCallException):
    """
    Exception for when walking a function input context.

    Raised in function calls.
    """


#
# Hook call exceptions
#
class HookCallException(TackleHookCallException):
    """
    Exception for an unknown field in a hook.

    Raised when field has been provided not declared in the hook type.
    """


class HookUnknownChdirException(TackleHookCallException):
    """
    Exception for a hook call with a chdir method.

    Raised when chdir is to an unknown directory.
    """


#
# Parser exceptions
#
class AppendMergeException(TackleParserException):
    """
    Excpetion when user tries to merge from for loop (ie hook --for [] --merge).

    Raised when handling empty blocks.
    """


class EmptyBlockException(TackleParserException):
    """
    Exception when a block (ie a compact hook without a hook type - key->: if:/for:)
    is declared without any contents.

    Raised when handling empty blocks.
    """


class UnknownArgumentException(TackleParserException):
    """
    Exception unknown argument when supplied via a hook call.

    Raised when a hook is called with an unknown argument in it's mapping (ie missing
    `_args` param in the hook definition.
    """


class NoInputOrParentTackleException(TackleParserException):
    """
    Exception when no input has been given, nor is there a parent tackle.

    Raised when a hook is called with an unknown argument in it's mapping (ie missing
    `_args` param in the hook definition.
    """


class UnknownTemplateVariableException(TackleParserException):
    """
    Exception an unknown templatable argument.

    Raised when rendering variables.
    """


class UnknownHookTypeException(TackleParserException):
    """
    Exception for unknown hook type.

    Raised when a hook type is not available from the providers.
    """


#
# Parser input exceptions
#
class EmptyTackleFileException(TackleParserInputException):
    """
    Exception when base tackle file is empty.

    Raised when calling files / providers.
    """


class UnknownSourceException(TackleParserInputException):
    """
    Exception for ambiguous source.

    Raised when tackle cannot determine which directory is the project
    template, e.g. more than one dir appears to be a template dir.
    """


#
# Util exceptions
#
class InvalidConfiguration(TackleException):
    """
    Exception for invalid configuration file.

    Raised if the global configuration file is not valid YAML or is
    badly constructed.
    """


# class UnknownRepoType(TackleException):
#     """
#     Exception for unknown repo types.
#
#     Raised if a repo's type cannot be determined.
#     """


class VCSNotInstalled(TackleException):
    """
    Exception when version control is unavailable.

    Raised if the version control system (git or hg) is not installed.
    """


class VersionNotFoundError(TackleException):
    """
    Exception when a version to a provider is given but does not exist.

    Raised if a version is specified but does not exist on the remote.
    """


class UnsupportedBaseFileTypeException(TackleException):
    """
    Exception for when a none json / yaml file are read

    Raised if the base file that is being called is not json / yaml.
    """


class ContextDecodingException(TackleException):
    """
    Exception for failed JSON decoding.

    Raised when a project's JSON context file can not be decoded.
    """


class RepositoryNotFound(TackleException):
    """
    Exception for missing repo.

    Raised when the specified tackle repository doesn't exist.
    """


class ContextFileNotFound(TackleException):
    """
    Exception for missing context file.

    Raised when the specified context file doesn't exist.
    """


class RepositoryCloneFailed(TackleException):
    """
    Exception for un-cloneable repo.

    Raised when a tackle template can't be cloned.
    """


class InvalidZipRepository(TackleException):
    """
    Exception for bad zip repo.

    Raised when the specified tackle repository isn't a valid
    Zip archive.
    """



"""Example."""
from tackle.models import BaseHook


class ThingHook(BaseHook):
    """Do stuff and things."""

    hook_type: str = 'thing'
    stuff: str

    def exec(self):
        print(self.stuff)
        return self.stuff



"""Tests `import` in the `tackle.providers.tackle.hooks.match` hook."""
from tackle import tackle


def test_provider_system_hook_match(change_dir):
    """Run the source and check that the hooks imported the demo module."""
    context = tackle('loop.yaml')
    assert context['matches'][0]
    assert len(context['matches'][6]) == 2


def test_provider_system_hook_match_cases(change_dir):
    """Run the source and check that the hooks imported the demo module."""
    context = tackle('cases.yaml')
    assert context['matched_dict'] == 'this'
    assert context['fallback_dict'] == 'foo'



from ruamel.yaml import YAML
from tackle import tackle


def test_parser_methods_merge(change_curdir_fixtures):
    yaml = YAML()
    with open('petstore.yaml') as f:
        expected_output = yaml.load(f)

    output = tackle('merge-petstore-compact.yaml')
    assert output == expected_output


def test_parser_methods_try(change_curdir_fixtures):
    """Use try which should not have any output"""
    output = tackle('method-try.yaml')
    assert output == {}


def test_parser_methods_except(change_curdir_fixtures):
    """Use try which should not have any output"""
    output = tackle('method-except.yaml')
    assert output['compact'] == 'foo'
    assert output['str'] == 'foo'
    assert output['dic']['stuff'] == '{{stuff}}'
    assert output['dict_render_block'] == {'stuff': '{{stuff}}'}
    assert output['stuff'] == 'things'
    assert output['listed']['hook_call'][1]['stuff'] == 'things'


def test_parser_methods_when(change_curdir_fixtures):
    """Use try which should not have any output"""
    output = tackle('method-when.yaml')
    assert 'expanded' not in output


def test_parser_methods_else_hooks(change_curdir_fixtures):
    """Use try which should not have any output"""
    output = tackle('method-else.yaml')
    assert output['compact'] == 'foo'
    assert output['str'] == 'foo'
    assert output['dict_render']['stuff'] == '{{stuff}}'
    assert output['str_render_block'] == 'things'
    assert output['dict_render_block'] == {'stuff': '{{stuff}}'}
    assert output['stuff'] == 'things'
    assert output['listed']['hook_call'][1]['stuff'] == 'things'



from tackle.exceptions import TackleException


class UndefinedVariableInTemplate(TackleException):
    """
    Exception for out-of-scope variables.

    Raised when a template uses a variable which is not defined in the
    context.
    """

    def __init__(self, message, error, context):
        """Exception for out-of-scope variables."""
        self.message = message
        self.error = error
        self.context = context

    def __str__(self):
        """Text representation of UndefinedVariableInTemplate."""
        return self.message



"""Test the input source part of the parser."""
import pytest

from tackle.exceptions import EmptyTackleFileException, UnknownArgumentException
from tackle.cli import main

INPUT_SOURCES = [
    ("empty.yaml", EmptyTackleFileException),
    ("empty-with-functions.yaml", EmptyTackleFileException),
    ("exceptions/out-of-range-arg.yaml", UnknownArgumentException),
]


@pytest.mark.parametrize("input_file,exception", INPUT_SOURCES)
def test_parser_raises_exceptions(change_curdir_fixtures, input_file, exception):
    """Test raising exceptions."""
    with pytest.raises(exception):
        main([input_file])



import tackle as tkl
from pydantic import SecretStr

from tackle.models import BaseHook, Field


class TackleHook(BaseHook):
    """Hook for calling external tackle providers."""

    hook_type: str = 'tackle'

    # fmt: off
    input_string: str = Field(
        None,
        description="The input can be one of repo, file path, directory with tackle.yaml, zip file, or if left blank parent tackle file.")
    checkout: str = Field(None,
                          description="The branch or version to checkout for repo type inputs_strings.")
    context_file: str = Field(None, description="The file to run inside a repo input.")
    extra_context: dict = Field(
        None,
        description="Any additional context to use when calling the hook. Like existing context.")
    context: dict = Field(None,
                          description="A context to use that overrides the current context.")
    password: SecretStr = Field(None, description="A password to use for repo inputs.")
    directory: str = Field(None,
                           description="The directory to run inside for repo inputs.")
    find_in_parent: bool = Field(
        False, description="Search for target in parent directory. Only relevant for "
                           "local targets.")

    override: dict = Field({}, description="A dictionary of keys to override.")
    # fmt: on

    args: list = ['input_string']
    _docs_order = 0

    def exec(self) -> dict:
        if self.context:
            existing_context = self.context
        else:
            existing_context = self.public_context.copy()

            if self.existing_context:
                existing_context.update(self.existing_context)

            if self.extra_context:
                existing_context.update(self.extra_context)

        output_context = tkl.main.tackle(
            self.input_string,
            checkout=self.checkout,
            password=self.password,
            directory=self.directory,
            calling_directory=self.calling_directory,
            calling_file=self.calling_file,
            # Evaluated
            existing_context=existing_context,
            # Implicit
            provider_hooks=self.provider_hooks,
            no_input=self.no_input,
            global_kwargs=self.override,
            find_in_parent=self.find_in_parent,
            verbose=self.verbose,
        )

        return output_context



"""Tests dict input objects for `tackle.providers.tackle.hooks.tackle` module."""
from tackle.main import tackle

import pytest
import shutil


@pytest.fixture()
def clean_outputs():
    """Remove the outputs dir."""
    yield
    shutil.rmtree('output')


# def test_provider_system_hook_tackle(change_dir):
# #     # TODO Build example repo
#     context = tackle('tackle.yaml', no_input=True)
#     assert context


def test_provider_tackle_local(change_dir):
    output = tackle('local.yaml', no_input=True)
    assert output['shell']['stuff'] == 'bing'


def test_provider_tackle_local_no_context(change_dir):
    output = tackle('local-no-context.yaml', no_input=True)
    assert output['shell']['foo'] == 'bar'


def test_provider_tackle_local_prior_context(change_dir):
    output = tackle('local-prior-context.yaml', no_input=True)
    assert output['shell']['foo'] == 'bar'


def test_provider_tackle_block_tackle(change_dir):
    output = tackle('block-tackle.yaml', no_input=True)
    assert output



import re
import ast
from jinja2 import meta
from jinja2.exceptions import UndefinedError
from inspect import signature
from typing import TYPE_CHECKING, Any
from pydantic import ValidationError

from tackle.special_vars import special_variables
from tackle.exceptions import UnknownTemplateVariableException

if TYPE_CHECKING:
    from tackle.models import Context


def wrap_braces_if_not_exist(value):
    """Wrap with braces if they don't exist."""
    if '{{' in value and '}}' in value:
        # Already templated
        return value
    return '{{' + value + '}}'


def wrap_jinja_braces(value):
    """Wrap a string with braces so it can be templated."""
    if isinstance(value, str):
        return wrap_braces_if_not_exist(value)
    # Nothing else can be wrapped
    return value


def render_variable(context: 'Context', raw: Any):
    """
    Render the raw input. Does recursion with dict and list inputs, otherwise renders
    string.

    :param raw: The value to be rendered.
    :return: The rendered value as literal type.
    """
    if raw is None:
        return None
    elif isinstance(raw, str):
        return render_string(context, raw)
    elif isinstance(raw, dict):
        return {
            render_string(context, k): render_variable(context, v)
            for k, v in raw.items()
        }
    elif isinstance(raw, list):
        return [render_variable(context, v) for v in raw]
    else:
        return raw


def render_string(context: 'Context', raw: str):
    """
    Render strings by first extracting renderable variables then build a render context
    from the public_context, then existing context, and last looks up special variables.
    After the value has been rendered it is returned as literal so as to preserve the
    original type of the value.

    :return: The literal value if the output is a string / list / dict / float / int
    """
    if '{{' not in raw:
        return raw

    template = context.env_.from_string(raw)
    # Extract variables
    variables = meta.find_undeclared_variables(context.env_.parse(raw))

    # Build a render context by inspecting the renderable variables
    render_context = {}
    unknown_variables = []
    for v in variables:
        # Variables in the current public_context take precedence
        if context.public_context and v in context.public_context:
            render_context.update({v: context.public_context[v]})
        elif context.private_context and v in context.private_context:
            render_context.update({v: context.private_context[v]})
        elif context.temporary_context and v in context.temporary_context:
            render_context.update({v: context.temporary_context[v]})
        elif context.existing_context and v in context.existing_context:
            render_context.update({v: context.existing_context[v]})
        elif v in special_variables:
            # If it is a special variable we need to check if the call requires
            # arguments, only context supported now.
            argments = list(signature(special_variables[v]).parameters)
            if len(argments) == 0:
                render_context.update({v: special_variables[v]()})
            elif 'context' in argments:
                render_context.update({v: special_variables[v](context)})
            else:
                raise ValueError("This should never happen.")
        else:
            unknown_variables.append(v)

    # Evaluate any hooks and insert into jinja environment globals
    if len(unknown_variables) != 0:
        # Unknown variables can be real unknown variables, preloaded jinja globals or
        # hooks which need to be inserted into the global env so that they can be called
        for i in unknown_variables:
            if i in context.provider_hooks:
                from tackle.models import LazyBaseFunction

                jinja_hook = context.provider_hooks[i]
                if isinstance(jinja_hook, LazyBaseFunction):
                    from tackle.parser import create_function_model

                    jinja_hook.dict().update(
                        {
                            'input_context': context.input_context,
                            'public_context': context.public_context,
                            'existing_context': context.existing_context,
                            'no_input': context.no_input,
                            'calling_directory': context.calling_directory,
                            'calling_file': context.calling_file,
                            'provider_hooks': context.provider_hooks,
                            'key_path': context.key_path,
                            'verbose': context.verbose,
                        }
                    )

                    # Need to instantiate the hook before passing the wrapped_exec
                    context.env_.globals[i] = create_function_model(
                        context, i, jinja_hook.dict()
                    )().wrapped_exec
                else:
                    context.env_.globals[i] = jinja_hook(
                        input_context=context.input_context,
                        public_context=context.public_context,
                        existing_context=context.existing_context,
                        no_input=context.no_input,
                        calling_directory=context.calling_directory,
                        calling_file=context.calling_file,
                        provider_hooks=context.provider_hooks,
                        key_path=context.key_path,
                        verbose=context.verbose,
                    ).wrapped_exec

    try:
        rendered_template = template.render(render_context)
        if rendered_template.startswith('<bound method BaseHook'):
            # Handle unknown variables that are the same as hook_types issues/55
            raise UndefinedError

        # Check for ambiguous globals like `namespace` tackle-box/issues/19
        match = re.search(r'\<class \'(.+?)\'>', rendered_template)
        if match:
            ambiguous_key = match.group(1).split('.')[-1].lower()
            if ambiguous_key in context.public_context:
                rendered_template = context.public_context[ambiguous_key]
            elif match.group(1) in context.existing_context:
                rendered_template = context.existing_context[ambiguous_key]

    except ValidationError as e:
        # For pydantic validation errors
        raise e

    except UndefinedError as e:
        if len(unknown_variables) != 0:
            raise UnknownTemplateVariableException(
                f"Variable{'s' if len(unknown_variables) != 1 else ''} {' '.join(unknown_variables)} unknown.",
                context=context,
            ) from None
        raise e

    try:
        # This will error on strings
        return ast.literal_eval(rendered_template)
    except (ValueError, SyntaxError):
        return rendered_template



import pytest

from tackle import exceptions, main


def test_should_raise_error_if_variable_does_not_exist(change_curdir_fixtures):
    with pytest.raises(exceptions.UnknownTemplateVariableException):
        main.tackle('unknown-variable.yaml')



import sys
from PyInquirer import prompt

try:
    from rich import pprint
except ImportError:
    from pprint import pprint

from tackle.models import BaseHook


class DebugHook(BaseHook):
    """Hook for debugging that prints the output context and pauses runtime."""

    hook_type: str = 'debug'

    def exec(self) -> None:
        """Run the hook."""
        print("Existing context")
        pprint(dict(self.existing_context))
        print("Local context")
        pprint(dict(self.public_context))
        if not self.no_input:
            question = {
                'type': 'confirm',
                'name': 'tmp',
                'message': 'CONTINUE',
            }
            response = prompt([question])

            # Catch keyboard exits with return an empty dict
            if response == {}:
                sys.exit(0)



# """
# This is the old yaml hook. Perhaps later it can be inspiration for a yaml hook that
# updates files in place? Could be useful to just point to a yaml file and combine the
# read and write in one operation with a swiss army knife of different filters...
# Maybe...
# """
#
# """YAML hook."""
# import logging
# import yaml
# import re
# import os
# from pydantic import Field
#
# from typing import Union, Dict, List, Any
#
# from tackle.models import BaseHook
# from tackle.utils import merge_configs
#
# logger = logging.getLogger(__name__)
#
#
# class YamlHook(BaseHook):
#     """
#     Hook for modifying a yaml in place (ie read, transform, and write back to the file
#     in one operation).
#     """
#     hook_type: str = 'yaml_in_place'
#
#     path: str = Field(..., description="The file path to put read or write to.")
#     remove: Union[List, str] = Field(
#         None, description="Parameter or regex to remove from list or dict."
#     )
#     contents: Union[Dict, List] = Field(
#         None, description="Supplied dictionary or list to write."
#     )
#     update: Dict = Field(
#         None,
#         description="Use the python `update` dict method on `contents` before writing",
#     )
#     filter: List = Field(None, description="List or string to values to.")
#     merge_dict: dict = Field(
#         None, description="Dict input that recursively overwrites the `contents`."
#     )
#     in_place: bool = Field(
#         False,
#         description="Boolean to read the contents of the `path` and then write after modifications.",
#     )
#     append_items: Union[Dict, str, List[Any]] = Field(
#         None, description="List to append to `append_key` key."
#     )
#     append_keys: Union[Dict, str, List[Any]] = Field(
#         None,
#         description="String or list of hierarchical keys to append item to. Defaults",
#     )
#     mode: str = Field(
#         None,
#         description="The mode that the file should write. Defaults to write 'w'. See https://docs.python.org/3/library/functions.html#open",
#     )
#     write: bool = Field(None, description="")
#
#     args: list = ['path', 'contents']
#
#     def __init__(self, **data: Any):
#         super().__init__(**data)
#         self.path = os.path.expanduser(os.path.expandvars(self.path))
#
#     def _load_contents(self):
#         if self.contents:
#             # We are writing. Context is provided
#             self.write = True
#             pass
#         if self.in_place:
#             # We are modifying in place. Context is read from path
#             self.write = True
#             with open(self.path, 'r') as f:
#                 self.contents = yaml.safe_load(f)
#         elif not self.contents:
#             # We are reading. Contents is read from path
#             self.write = False
#             mode = self.mode or 'r'
#             with open(self.path, mode) as f:
#                 self.contents = yaml.safe_load(f)
#
#     def _append_each_item(self, append_item):
#         if isinstance(self.append_keys, str):
#             self.contents[self.append_keys].append(append_item)
#         elif isinstance(self.append_keys, list):
#             for k in self.append_keys[:-1]:
#                 self.contents = self.contents.setdefault(k, {})
#             self.contents[self.append_keys[-1]].append(append_item)
#         else:
#             self.contents.append(append_item)
#
#     def _remove_from_contents(self, regex):
#         if isinstance(self.contents, list):
#             self.contents = [i for i in self.contents if not re.search(regex, i)]
#         if isinstance(self.contents, dict):
#             for k in list(self.contents.keys()):
#                 if re.search(regex, k):
#                     self.contents.pop(k)
#
#     def _modify_dicts(self):
#         if self.remove:
#             if isinstance(self.remove, str):
#                 self._remove_from_contents(self.remove)
#
#             if isinstance(self.remove, list):
#                 for i in self.remove:
#                     self._remove_from_contents(i)
#
#         if self.filter:
#             if isinstance(self.contents, dict):
#                 self.contents = {
#                     k: v for (k, v) in self.contents.items() if k in self.filter
#                 }
#
#         if self.update:
#             self.contents.update(self.update)
#
#         if self.merge_dict:
#             self.contents = merge_configs(self.contents, self.merge_dict)
#
#         if self.append_items:
#             if isinstance(self.append_items, str) or isinstance(
#                 self.append_items, dict
#             ):
#                 self._append_each_item(self.append_items)
#             elif isinstance(self.append_items, list):
#                 for i in self.append_items:
#                     self._append_each_item(i)
#
#     def exec(self) -> Union[str, dict]:
#         # Load the path into contents unless it already exists
#         self._load_contents()
#         # Run all the modifiers
#         self._modify_dicts()
#
#         if self.write:
#             mode = self.mode or 'w'
#             with open(self.path, mode) as f:
#                 yaml.dump(self.contents, f)
#                 return self.contents
#         else:
#             # Read operation, just return contents
#             return self.contents



import pytest
from ruamel.yaml import YAML

from tackle import tackle
from tackle.exceptions import FunctionCallException

FIXTURES = [
    ('call.yaml', 'call-output.yaml'),
    ('return.yaml', 'return-output.yaml'),
]


@pytest.mark.parametrize("fixture,expected_output", FIXTURES)
def test_function_model_extraction(change_curdir_fixtures, fixture, expected_output):
    yaml = YAML()
    with open(expected_output) as f:
        expected_output_dict = yaml.load(f)

    output = tackle(fixture)
    assert output == expected_output_dict


EXCEPTION_FIXTURES = [('return-str-not-found.yaml', FunctionCallException)]


@pytest.mark.parametrize("fixture,exception", EXCEPTION_FIXTURES)
def test_function_raises_exceptions(change_curdir_fixtures, fixture, exception):
    with pytest.raises(exception):
        tackle(fixture)



"""Test the input source part of the parser."""
# import pytest
# import shutil

from tackle.parser import update_source
from tackle.models import Context

# INPUT_SOURCES = [
#     ("github.com/robcxyz/tackle-demo", 'my')
#     # ("robcxyz/full-stack-fastapi-postgresql", 'base-project')
# ]
#
#
# @pytest.mark.parametrize("input_string,output_dir", INPUT_SOURCES)
# def test_parser_update_source(change_curdir_fixtures, input_string, output_dir):
#     """Test various inputs."""
#     context = Context(input_string=input_string, no_input=True)
#     update_source(context)
#     shutil.rmtree(output_dir)


def test_parser_update_source_key_to_parent(chdir, mocker):
    """
    Test that when an input string is not a file/provider source that we correctly
    traverse to the nearest tackle file and run a key within it.
    """
    chdir('fixtures/input-key/child')
    context = Context(input_string="do_things", no_input=True)
    update_source(context)
    print()



"""Verify hook argument related items."""
import pytest
from tackle.utils.command import unpack_args_kwargs_string
from tackle import tackle

FIXTURES = [
    ("this --if \"expanded == 'that'\"", ["this"], {"if": "expanded == 'that'"}, []),
    (
        "this that --if \"expanded == 'that'\"",
        ["this", 'that'],
        {"if": "expanded == 'that'"},
        [],
    ),
]


@pytest.mark.parametrize("input_string,args,kwargs,flags", FIXTURES)
def test_unpack_input_string(input_string, args, kwargs, flags):
    """Verify the hook arguments."""
    args_out, kwargs_out, flags_out = unpack_args_kwargs_string(input_string)
    assert args_out == args
    assert kwargs_out == kwargs
    assert flags_out == flags


def test_parser_multiple_args(change_curdir_fixtures):
    """Test input args."""
    output = tackle('args.yaml', no_input=True)
    assert output['two_args'] == 'foo bar'
    assert output['three_args'] == 'foo bar baz'


def test_parser_tackle_in_tackle(change_curdir_fixtures):
    """Test input args."""
    output = tackle('outer-tackle.yaml', no_input=True)
    assert output['outer']['foo_items'] == ['bar', 'baz']


def test_parser_tackle_in_tackle_arg(change_curdir_fixtures):
    """Test input args."""
    output = tackle('outer-tackle-arg.yaml', no_input=True)
    assert output['outer']['foo_items'] == ['bar', 'baz']


def test_parser_render_hook_input(change_curdir_fixtures):
    """
    Show that when the first argument is a renderable that you just render that with the
    context and logic.
    """
    output = tackle('var-hook.yaml', no_input=True)
    for k, v in output.items():
        if k.startswith('stuff'):
            assert v == 'things'


# TODO: When help is done
# @pytest.mark.parametrize("input_string", ['do_stuff', 'tackle.yaml do_stuff'])
# def test_parser_call_key_on_args(chdir_fixture, input_string):
#     """Validate that when an additional argument is supplied that the key is called."""
#     chdir_fixture('help')
#     output = tackle(input_string)
#     assert output['do_stuff'] == 'Doing things...'
#     assert output['do_things'] != 'Doing stuff...'


# @pytest.mark.parametrize(
#     "input_string",
#     [
#         # 'do_stuff',
#         # 'file.yaml do_stuff do_things'
#         'file.yaml do_stuff_compact do_things and other things'
#         # 'file.yaml do_stuff_compact_block print_this and print all this to'
#     ],
# )
# def test_parser_call_key_on_args_cli(chdir_fixture, input_string):
#     chdir_fixture('help')
#     main(input_string.split())



import pytest
import os
from tackle.main import tackle
import subprocess
import sys


@pytest.fixture()
def temporary_uninstall():
    """Fixture to uninstall a package and install it after the test."""

    def f(package):
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "--quiet",
                "--disable-pip-version-check",
                "-y",
                package,
            ]
        )

    return f


def test_parser_provider_import_installs_requirements(
    change_curdir_fixtures, temporary_uninstall
):
    """Validate that if a package is missing, that it will be installed and usable."""
    temporary_uninstall('requests')
    tackle('test-install-dep.yaml')


def test_parser_hooks_raises_error_on_unknown_hook_type(change_curdir_fixtures):
    """Verify raising error.

    Verify that the hook parser raises the right error when the hook type is
    not in the providers.
    """
    from tackle.exceptions import UnknownHookTypeException

    with pytest.raises(UnknownHookTypeException):
        tackle('unknown-hook-type.yaml')


def test_parser_provider_hook_add(change_curdir_fixtures):
    """Validate adding providers.

    Validate that you can give a `__provider` key to point to
    additional providers and make them available as a type.
    """
    os.chdir('test-provider')
    o = tackle('context_provider.yaml')
    assert o['things'] == 'bar'


def test_parser_provider_hook_add_list(change_curdir_fixtures):
    """Validate that hooks are imported when called from the same directory."""
    o = tackle('context-provider-list.yaml')
    assert o['things'] == 'bar'
    assert o['stuff'] == 'bar'


def test_providers_local_hooks_dir(chdir_fixture):
    """
    Check that when you call from a child dir that you import the hooks from the dir
     with the file being called.
    """
    chdir_fixture(os.path.join('child', 'dir'))
    o = tackle()
    assert o



"""
Utils for handling command arguments along with unpacking hook arguments. Used in cli
to unpack input args/kwargs and in parser to unpack hook calls -
i.e. key->: hook arg --kwarg thing
"""
import ast
import re


def strip_dashes(raw_arg: str) -> str:
    """Remove dashes prepended to string for arg splitting."""
    while raw_arg.startswith('-'):
        raw_arg = raw_arg[1:]
    return raw_arg


def split_input_string(input_string: str) -> list:
    """
    Split first on whitespace then regex each item to qualify if it needs to be
    intepreted as literal (ast).
    """
    # Inspired from https://stackoverflow.com/a/524796/12642712
    # input_list = [p for p in re.split("( |(?<!\{|\[)\\\"(?!\,|\:).*?\\\"(?!\}|\])|'.*?')", input_string) if p.strip()]

    # Split on whitespace
    # When quotes are preceded by `,:{[(=`, ignore
    # Don't split between quotes with `,:]})`
    # Otherwise split on quotes
    # Repeated for single quotes
    input_list = [
        p
        for p in re.split(
            "( |(?<!,|\:|\(|\{|\[|=)\\\"(?!\,|\:|\)|\}|\]).*?\\\"(?!\}|\])|(?<!\,|\:|\(|\{|\[|=)'(?!,|\:|\)|\}|\]).*?'(?!\}|\]))",  # noqa
            input_string,
        )
        if p.strip()
    ]

    output = []
    for i in input_list:
        try:
            output.append(ast.literal_eval(i))
        except (ValueError, SyntaxError):
            output.append(i)
    return output


def unpack_args_kwargs_string(input_string: str) -> (list, dict, list):
    """Split up based on whitespace input args and pass to unpack_args_kwargs_list."""
    # input_list = shlex.split(input_string)
    input_list = split_input_string(input_string)
    return unpack_args_kwargs_list(input_list)


def assert_if_flag(arg: str):
    FLAG_REGEX = re.compile(
        r"""^[\-|\-\-]+[a-zA-Z0-9]""",
        re.VERBOSE,
    )
    return bool(FLAG_REGEX.match(arg))


def unpack_args_kwargs_list(input_list: list) -> (list, dict, list):
    """
    Take the input template and unpack the args and kwargs if they exist.
    Updates the command_args and command_kwargs with a list of strings and
    list of dicts respectively.
    """
    input_list_length = len(input_list)
    args = []
    kwargs = {}
    flags = []

    i = 0
    while i < input_list_length:
        raw_arg = input_list[i]

        # Look ahead for if the last item is a flag
        if i + 1 < input_list_length:
            next_raw_arg = input_list[i + 1]
        else:
            # Hack for asserting if a flag - two items starting with a dash or last item
            next_raw_arg = "--hack"

        if isinstance(raw_arg, str):
            # if raw_arg.startswith('-'):
            if assert_if_flag(raw_arg):
                if isinstance(next_raw_arg, str):
                    # if next_raw_arg.startswith('-'):
                    if assert_if_flag(next_raw_arg):
                        # Field is a flag
                        flags.append(strip_dashes(raw_arg))
                    else:
                        # Field is a kwarg
                        kwargs.update({strip_dashes(raw_arg): input_list[i + 1]})
                        i += 1
                else:
                    # Field is a kwarg
                    kwargs.update({strip_dashes(raw_arg): input_list[i + 1]})
                    i += 1
            else:
                # Field is an argument
                args.append(raw_arg)
        else:
            # Field is an argument
            args.append(raw_arg)

        i += 1
    return args, kwargs, flags



import os
import re
import logging
from pathlib import Path
from functools import partialmethod
from typing import Type, Any, Union
from pydantic import Field, create_model
from pydantic.main import ModelMetaclass
from collections import OrderedDict

from tackle.render import render_variable, wrap_jinja_braces
from tackle.utils.dicts import (
    nested_get,
    nested_delete,
    nested_set,
    encode_list_index,
    decode_list_index,
    set_key,
    get_target_and_key,
    smush_key_path,
)
from tackle.utils.command import unpack_args_kwargs_string
from tackle.utils.vcs import get_repo_source
from tackle.utils.files import read_config_file
from tackle.utils.paths import (
    work_in,
    is_repo_url,
    is_directory_with_tackle,
    is_file,
    find_tackle_file,
    find_nearest_tackle_file,
    find_in_parent,
)
from tackle.utils.zipfile import unzip
from tackle.models import (
    Context,
    BaseHook,
    LazyImportHook,
    BaseFunction,
    FunctionInput,
    LazyBaseFunction,
)
from tackle.exceptions import (
    HookCallException,
    UnknownHookTypeException,
    UnknownArgumentException,
    UnknownSourceException,
    EmptyTackleFileException,
    EmptyBlockException,
    FunctionCallException,
    HookUnknownChdirException,
    AppendMergeException,
)
from tackle.settings import settings

logger = logging.getLogger(__name__)


def get_hook(hook_type, context: 'Context'):
    """
    Get the hook from available providers and install hook requirements if they haven't
    been already.

    This function does the following to return the hook:
    1. Check if hook hasn't been imported already
    2. Check if the hook has been declared in a provider's __init__.py's
    `hook_types` field.
    3. Try to import it then fall back on installing the requirements.txt file
    """
    h = context.provider_hooks.get(hook_type, None)
    if h is None:
        # Show this without verbose:
        available_hooks = (
            "Run the application with `--verbose` to see available hook types."
        )
        if context.verbose:
            available_hooks = 'Available hooks = ' + ' '.join(
                [str(i) for i in context.provider_hooks.keys()]
            )
        raise UnknownHookTypeException(
            f"The hook type=\"{hook_type}\" is not available in the providers. "
            + available_hooks
        )

    # LazyImportHook in hook ref when declared in provider __init__.hook_types
    if isinstance(h, LazyImportHook):
        # Install the requirements which will convert all the provider
        context.provider_hooks.import_with_fallback_install(
            mod_name=h.mod_name,
            path=h.hooks_path,
        )
        h = context.provider_hooks[hook_type]
    return h


def evaluate_for(hook_dict: dict, Hook: ModelMetaclass, context: 'Context'):
    """Run the parse_hook function in a loop and return a list of outputs."""
    loop_targets = render_variable(context, wrap_jinja_braces(hook_dict['for']))
    hook_dict.pop('for')

    # Need add an empty list in the value so we have something to append to
    target_context, key_path = get_target_and_key(context)
    nested_set(target_context, key_path, [])

    if len(context.key_path_block) != 0:
        tmp_key_path = context.key_path[
            -(len(context.key_path) - len(context.key_path_block)) :
        ]
        if context.temporary_context is None:
            context.temporary_context = {} if isinstance(tmp_key_path[0], str) else []
        tmp_key_path = [i for i in tmp_key_path if i not in ('->', '_>')]
        nested_set(context.temporary_context, tmp_key_path, [])

    if len(loop_targets) == 0:
        return

    for i, l in (
        enumerate(loop_targets)
        if not render_variable(context, hook_dict.get('reverse', None))
        else reversed(list(enumerate(loop_targets)))
    ):
        if context.existing_context is None:
            context.existing_context = {}
        # Create temporary variables in the context to be used in the loop.
        context.existing_context.update({'index': i, 'item': l})
        # Append the index to the keypath
        context.key_path.append(encode_list_index(i))

        # TODO: Do we need to parse a copy of the hook?
        parse_hook(hook_dict.copy(), Hook, context, append_hook_value=True)
        context.key_path.pop()

    # Remove temp variables
    try:
        context.existing_context.pop('item')
        context.existing_context.pop('index')
    except KeyError:
        pass


def evaluate_if(hook_dict: dict, context: 'Context', append_hook_value: bool) -> bool:
    """Evaluate the if/when condition and return bool."""
    if hook_dict.get('when', None) is not None:
        result = render_variable(context, wrap_jinja_braces(hook_dict['when']))
        hook_dict.pop('when')
        return result
    if hook_dict.get('for', None) is not None and not append_hook_value:
        # We qualify `if` conditions within for loop logic
        return True
    if hook_dict.get('if', None) is None:
        return True

    return render_variable(context, wrap_jinja_braces(hook_dict['if']))


def merge_block_output(
    hook_output_value: Any,
    context: Context,
    append_hook_value: bool = False,
):
    """
    Block hooks have already written to the output dict so to merge, need to take the
     keys from the key path and move them up one level.
    """
    if append_hook_value:
        raise AppendMergeException("Can't merge from for loop.", context=context)

    target_context, key_path = get_target_and_key(context)
    indexed_block_output = nested_get(element=hook_output_value, keys=key_path)
    for k, v in indexed_block_output.items():
        element, key_path = get_target_and_key(context)
        nested_set(
            element=target_context,
            keys=key_path[:-1] + [k],
            value=v,
        )
    nested_delete(element=target_context, keys=key_path)


def merge_output(
    hook_output_value: Any,
    context: Context,
    append_hook_value: bool = False,
):
    """Merge the contents into it's top level set of keys."""
    if append_hook_value:
        raise AppendMergeException("Can't merge from for loop.")

    if context.key_path[-1] in ('->', '_>'):
        # Expanded key - Remove parent key from key path
        key_path = context.key_path[:-2] + [context.key_path[-1]]
    else:
        # Compact key
        key_path = context.key_path[:-1] + [context.key_path[-1][-2:]]

    # Can't merge into top level keys without merging k/v individually
    if len(key_path) == 1:
        # This is only valid for dict output
        if isinstance(hook_output_value, (dict, OrderedDict)):
            for k, v in hook_output_value.items():
                set_key(context=context, value=v, key_path=[k] + key_path)
            return
        else:
            raise HookCallException(
                "Can't merge non maps into top level keys.",
            )
    else:
        set_key(context=context, value=hook_output_value, key_path=key_path)


def run_hook_in_dir(hook: Type[BaseHook]) -> Any:
    if hook.chdir:
        path = os.path.abspath(os.path.expanduser(hook.chdir))
        if os.path.isdir(path):
            # Use contextlib to switch dirs
            with work_in(os.path.abspath(os.path.expanduser(hook.chdir))):
                return hook.exec()
        else:
            raise HookUnknownChdirException(
                f"The specified path='{path}' to change to was not found.",
                hook=hook,
            )
    else:
        return hook.exec()


def render_hook_vars(hook_dict: dict, Hook: ModelMetaclass, context: 'Context'):
    """Render the hook variables."""
    for key, value in list(hook_dict.items()):
        if key in Hook.Config.alias_to_fields:
            # Skip any keys used in logic as these are evaluated / rendered separately
            continue

        if key in Hook._render_exclude or key in Hook._render_exclude_default:
            # Skip anything that has been marked excluded. Needed for things like block
            # hooks which will have templating within the inputs
            continue

        if isinstance(value, str):
            # Check Hook private vars for rendering by default which wraps bare strings
            if key in Hook._render_by_default:
                hook_dict[key] = render_variable(context, wrap_jinja_braces(value))

            # TODO: When we build our own custom Field function then this will change
            elif 'render_by_default' in Hook.__fields__[key].field_info.extra:
                hook_dict[key] = render_variable(context, wrap_jinja_braces(value))

            elif '{{' in value and '}}' in value:
                hook_dict[key] = render_variable(context, value)

        elif isinstance(value, (list, dict)):
            hook_dict[key] = render_variable(context, value)


def parse_sub_context(
    context: 'Context', hook_dict: dict, target: str, append_hook_value: bool = None
):
    """
    Reparse a subcontext as in the case with `else` and `except` where you have to
    handle the negative side of the either `if` or `try`. Works on both looped and
    normal runs by checking the last item in the key path. Then overwrites the input
    with a new context.
    """
    if isinstance(hook_dict[target], str):
        set_key(
            context=context,
            value=render_variable(context, hook_dict[target]),
            append_hook_value=append_hook_value,
        )
        return

    indexed_key_path = context.key_path[
        -(len(context.key_path) - len(context.key_path_block)) :
    ]

    if isinstance(indexed_key_path[-1], bytes):
        # We are in a for loop so we need
        input_dict = nested_get(
            element=context.input_context,
            keys=indexed_key_path[:-3],
        )
        updated_item = [
            hook_dict[target] if i == decode_list_index(context.key_path[-1]) else None
            for i in range(decode_list_index(context.key_path[-1]) + 1)
        ]
        input_dict[indexed_key_path[-3]] = updated_item
        walk_sync(
            context,
            element=input_dict[indexed_key_path[-3]][
                decode_list_index(context.key_path[-1])
            ],
        )

    else:
        input_dict = nested_get(
            element=context.input_context,
            keys=indexed_key_path[:-2],
        )
        arrow = context.key_path[-1]
        input_dict[indexed_key_path[-2]] = {arrow: 'block', 'items': hook_dict[target]}
        walk_sync(context, element=input_dict[indexed_key_path[-2]])


def parse_hook(
    hook_dict, Hook: ModelMetaclass, context: 'Context', append_hook_value: bool = None
):
    """Parse input dict for loop and when logic and calls hooks."""
    if evaluate_if(hook_dict, context, append_hook_value):

        if 'for' in hook_dict:
            # This runs the current function in a loop with `append_hook_value` set so
            # that keys are appended in the loop.
            evaluate_for(hook_dict, Hook, context)
            return

        else:
            # Render the remaining hook variables
            render_hook_vars(hook_dict, Hook, context)

            hook = Hook(
                **hook_dict,
                input_context=context.input_context,
                public_context=context.public_context,
                private_context=context.private_context,
                temporary_context=context.temporary_context,
                existing_context=context.existing_context,
                no_input=context.no_input,
                calling_directory=context.calling_directory,
                calling_file=context.calling_file,
                provider_hooks=context.provider_hooks,
                key_path=context.key_path,
                verbose=context.verbose,
                env_=context.env_,
                is_hook_call=True,
            )
            # Main exec logic
            if hook.try_:
                try:
                    hook_output_value = run_hook_in_dir(hook)
                except Exception as e:
                    if hook.verbose:
                        print(e)
                    if hook.except_:
                        # TODO: Use generalized function same with `else`
                        # raise NotImplementedError
                        parse_sub_context(context, hook_dict, target='except')
                    return
            else:
                # Normal hook run
                hook_output_value = run_hook_in_dir(hook)

            if hook.skip_output:
                if hook.merge:
                    merge_block_output(
                        hook_output_value=hook_output_value,
                        context=context,
                        append_hook_value=append_hook_value,
                    )
                return
            elif hook.merge:
                merge_output(
                    hook_output_value=hook_output_value,
                    context=context,
                    append_hook_value=append_hook_value,
                )
            else:
                set_key(
                    context=context,
                    value=hook_output_value,
                    append_hook_value=append_hook_value,
                )

    elif 'else' in hook_dict:
        parse_sub_context(context, hook_dict, target='else')


def evaluate_args(args: list, hook_dict: dict, Hook: Type[BaseHook]):
    """
    Associate hook arguments provided in the call with hook attributes. Parses the
    hook's `_args` attribute to know how to map arguments are mapped to where and
    deal with rendering by default.

    TODO: This needs to be re-thought. Right now we parse the inputs without regard
     for the types of the argument mapping. What could be better is if we know the types
     of the arg mapping ahead of time and they try to assemble the most logical mapping
     afterwards. So if the mapping consists of a [str, list], then the if the first
     args are strs then we can ignore the list part. Right now it would just join all
     the strings together if they are part of last arg mapping.
    TODO: Improve error handling.
    Solutions:
    - First try to infer type from arg
        - Single types then into unions?
    - If type cannot be infered (ie Any) then do ast as literal
    """
    hook_args: list = Hook.__fields__['args'].default
    for i, v in enumerate(args):
        # Iterate over the input args
        if i + 1 == len(hook_args):
            # We are at the last argument mapping so we need to join the remaining
            # arguments as a single string if it is not a list of another map.
            if not isinstance(args[i], (str, float)):
                # Catch list dict and ints - strings floats and bytes caught later
                value = args[i]
            elif Hook.__fields__[hook_args[i]].type_ in (str, float, int):
                # Was parsed on spaces so reconstructed.
                value = ' '.join(args[i:])
            # fmt: skip
            elif isinstance(Hook.__fields__[hook_args[i]], list):
                # If list then all the remaining items
                value = args[i:]
            elif isinstance(v, (str, float, int)):
                # Make assumption the rest of the args can be reconstructed as above
                value = ' '.join(args[i:])
            else:
                # Only thing left is a dict
                if len(args[i:]) > 1:
                    raise ValueError(
                        f"Can't specify multiple arguments for map argument "
                        f"{Hook.__fields__[hook_args[i]]}."
                    )
                value = args[i]
            hook_dict[hook_args[i]] = value
            return
        else:
            # The hooks arguments are indexed
            try:
                hook_dict[hook_args[i]] = v
            except IndexError:
                raise UnknownArgumentException(f"Unknown argument {i}.")


def handle_leading_brackets(args) -> list:
    """
    Handler for cases where we have a hook call with a renderable string as the first
    argument which we rewrite as a var hook. For instance `foo->: foo-{{ bar }}-baz`
    would be rewritten as `foo->: var foo-{{bar}}-baz`.
    """
    if '{{' in args[0]:
        # We split up the string before based on whitespace so eval individually
        if '}}' in args[0]:
            # This is single templatable string -> key->: "{{this}}" => args: ['this']
            args.insert(0, 'var')
        else:
            # Situation where we have key->: "{{ this }}" => args: ['{{', 'this' '}}']
            for i in range(1, len(args)):
                if '}}' in args[i]:
                    joined_template = ' '.join(args[: (i + 1)])
                    other_args = args[(i + 1) :]
                    args = ['var'] + [joined_template] + other_args
                    break
    return args


def run_hook(context: 'Context'):
    """
    Run the hook by qualifying the input argument and matching the input params with the
    with the hook's `_args` which are then overlayed into a hook kwargs. Also interprets
    special cases where you have a string or list input of renderable variables.
    """
    if isinstance(context.input_string, str):
        args, kwargs, flags = unpack_args_kwargs_string(context.input_string)
        args = handle_leading_brackets(args)
        first_arg = args[0]
        # Remove first args it will be consumed and no longer relevant
        args.pop(0)

    else:
        # Rare case when an arrow is used to indicate rendering of a list.
        # Only qualified when input is of form `key->: [{{var}},{{var}},...]
        # In this case we need to set the key as an empty list
        set_key(context, value=[])

        # Iterate over values appending rendered values. Rendered values can be any type
        for i, v in enumerate(context.input_string):
            set_key(
                context=context,
                value=render_variable(context, v),
                key_path=context.key_path + [encode_list_index(i)],
            )
        return

    # Look up the hook from the imported providers
    Hook = get_hook(first_arg, context)

    if Hook is None:
        raise UnknownHookTypeException(
            f"Hook type-> \"{first_arg}\" unknown.", context=context
        )

    # TODO: Enrich lazy functions
    #  Might want to instead track lazy
    if isinstance(Hook, LazyBaseFunction):
        Hook = create_function_model(context, first_arg, Hook.dict())

    if context.key_path[-1] in ('->', '_>'):
        # We have an expanded or mixed (with args) hook expression and so there will be
        # additional properties in adjacent keys. Trim key_path_block for blocks
        hook_dict = nested_get(
            context.input_context,
            smush_key_path(context.key_path[:-1][len(context.key_path_block) :]),
        ).copy()
        # Need to replace arrow keys as for the time being (pydantic 1.8.2) - multiple
        # aliases for the same field (type) can't be specified so doing this hack
        if '->' in hook_dict:
            hook_dict.pop('->')
        else:
            hook_dict.pop('_>')
    else:
        # Hook is a compact expression - Can only be a string
        hook_dict = {}
    hook_dict['hook_type'] = first_arg

    # Associate hook arguments provided in the call with hook attributes
    evaluate_args(args, hook_dict, Hook)
    # Add any kwargs
    for k, v in kwargs.items():
        hook_dict[k] = v
    for i in flags:
        hook_dict[i] = True

    # Main parser
    parse_hook(hook_dict, Hook, context)


def handle_empty_blocks(context: 'Context', block_value):
    """
    Handle keys appended with arrows and interpret them as `block` hooks. Value is
    re-written over with a `block` hook to support the following syntax.

    a-key->:
      if: stuff == 'things'
      for: a_list
      foo->: print ...
      bar->: print ...

    to

    a-key:
      ->: block
      if: stuff == 'things'
      for: a_list
      items:
        foo->: print ...
        bar->: print ...

    :param context:
    :param block_key:
    :param block_value:
    :return:
    """
    # Break up key paths
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]
    # Handle embedded blocks which need to have their key paths adjusted
    key_path = (base_key_path + new_key)[len(context.key_path_block) :]
    arrow = [context.key_path[-1][-2:]]

    indexed_key_path = context.key_path[
        -(len(context.key_path) - len(context.key_path_block)) :
    ]
    input_dict = nested_get(
        element=context.input_context,
        keys=indexed_key_path[:-1],
    )
    value = input_dict[indexed_key_path[-1]]

    for k, v in list(input_dict.items()):
        if k == indexed_key_path[-1]:
            nested_set(context.input_context, key_path, {arrow[0]: 'block'})
            nested_delete(context.input_context, indexed_key_path)
        else:
            context.input_context[k] = context.input_context.pop(k)

    # Iterate through the block keys except for the reserved keys like `for` or `if`
    aliases = [v.alias for _, v in BaseHook.__fields__.items()] + ['->', '_>']
    for k, v in value.copy().items():
        if k in aliases:
            nested_set(
                element=context.input_context,
                keys=key_path + [k],
                value=v,
            )
        else:
            # Set the keys under the `items` key per the block hook's input
            nested_set(
                element=context.input_context,
                keys=key_path + ['items', k],
                value=v,
            )


def compact_hook_call_macro(context: 'Context', element: str) -> dict:
    """
    Rewrite the input_context with an expanded expression on the called compact key.
     Returns the string element as a dict with the key as the arrow and element as
     value.
    """
    # TODO: Clean this up
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]
    key_path = (base_key_path + new_key)[len(context.key_path_block) :]
    arrow = context.key_path[-1][-2:]

    extra_keys = len(context.key_path) - len(context.key_path_block)
    old_key_path = context.key_path[-extra_keys:]

    value = nested_get(
        element=context.input_context,
        keys=smush_key_path(old_key_path)[:-1],
    )

    replacement = {context.key_path[-1]: new_key[0]}
    for k, v in list(value.items()):
        value[replacement.get(k, k)] = (
            value.pop(k) if k != context.key_path[-1] else {arrow: value.pop(k)}
        )

    # Reset the key_path without arrow
    context.key_path = context.key_path_block + key_path

    return {arrow: element}


def list_to_var_macro(context: 'Context', element: list) -> dict:
    """
    Convert arrow keys with a list as the value to `var` hooks via a re-write to the
    input.
    """
    # TODO: Convert this to a block. Issue is that keys are not rendered by default so
    #  when str items in a list are parsed, they are not rendered by default. Should
    #  have some validator or something on block to render str items in a list.
    base_key_path = context.key_path[:-1]
    new_key = [context.key_path[-1][:-2]]
    old_key_path = context.key_path[len(context.key_path_block) :]
    arrow = context.key_path[-1][-2:]

    _, key_path = get_target_and_key(context)
    if isinstance(context.input_context, dict):
        nested_set(
            element=context.input_context,
            keys=base_key_path[-(len(base_key_path) - len(context.key_path_block)) :]
            + new_key,
            value={arrow: 'var', 'input': element},
        )
        # Remove the old key
        nested_delete(context.input_context, old_key_path)
        context.key_path = base_key_path + new_key

    else:
        context.input_context = {arrow: 'var', 'input': element}
        context.key_path = base_key_path

    return {arrow: 'var'}


def walk_sync(context: 'Context', element):
    """
    Traverse an object looking for hook calls and running those hooks. Here we are
    keeping track of which keys are traversed in a list called `key_path` with strings
    as dict keys and byte encoded integers for list indexes.
    """
    if len(context.key_path) != 0:
        # Handle compact expressions - ie key->: hook_type args
        # if context.key_path[-1][-2:] in ('->', '_>'):
        ending = context.key_path[-1][-2:]
        if ending in ('->', '_>'):
            if isinstance(element, str):
                element = compact_hook_call_macro(context, element)
            elif isinstance(element, list):
                element = list_to_var_macro(context, element)

    if isinstance(element, dict):
        # Handle expanded expressions - ie key:\n  ->: hook_type args
        if '->' in element.keys():
            # Public hook calls
            context.key_path.append('->')
            context.input_string = element['->']
            run_hook(context)
            context.key_path.pop()
            return
        elif '_>' in element.keys():
            # Private hook calls
            context.key_path.append('_>')
            context.input_string = element['_>']
            run_hook(context)
            context.key_path.pop()
            return
        elif element == {}:
            set_key(context=context, value={})
            return

        for k, v in element.copy().items():
            context.key_path.append(k)
            # Special case where we have an empty hook, expanded or compact
            if k[-2:] in ('->', '_>') and (v is None or isinstance(v, dict)):
                # Here we re-write the input to turn empty hooks into block hooks
                handle_empty_blocks(context, v)
                context.key_path[-1] = k[:-2]
                value = nested_get(
                    element=context.input_context,
                    keys=context.key_path[
                        -(len(context.key_path) - len(context.key_path_block)) :
                    ],
                )
                # Finally check if the `items` key exists in the input_context.  If not
                #  then we have an empty hook which will cause an ambiguous
                #  ValidationError for missing field
                if 'items' not in value:
                    raise EmptyBlockException("Empty block hook.", context=context)

                walk_sync(context, value)
                context.key_path.pop()
            else:
                # Recurse
                walk_sync(context, v)
                context.key_path.pop()
    # Non-hook calls recurse through inputs
    elif isinstance(element, list):
        # Handle empty lists
        if len(element) == 0:
            set_key(context=context, value=element)
        else:
            for i, v in enumerate(element.copy()):
                context.key_path.append(encode_list_index(i))
                walk_sync(context, v)
                context.key_path.pop()
    else:
        set_key(context=context, value=element)


def run_handler(context, handler_key, handler_value):
    """
    Run a pre/post execution handlers which are either hooks or functions.

    NOTE: This is an experimental feature and may change.
    """
    if handler_key in context.functions:
        """Run functions"""
        function = context.functions[handler_key]
        context.input_context = function.exec
        walk_sync(context, function.exec.copy())

    elif get_hook(handler_key, context):
        Hook = get_hook(handler_key, context)
        hook = Hook(
            **handler_value,
            input_context=context.input_context,
            public_context=context.public_context,
            no_input=context.no_input,
            providers=context.providers,
        )
        hook.call()
    else:
        raise UnknownHookTypeException(
            f"Unknown hook type={handler_key}.", context=context
        )


def update_input_context_with_kwargs(context: 'Context', kwargs: dict):
    """
    Update the input dict with kwargs which in this context are treated as overriding
    the keys. Takes into account if the key is a hook and replaces that.
    """
    for k, v in kwargs.items():
        if k in context.input_context:
            context.input_context.update({k: v})
        elif f"{k}->" in context.input_context:
            # Replace the keys and value in the same position it was in
            context.input_context = {
                key if key != f"{k}->" else k: value if key != f"{k}->" else v
                for key, value in context.input_context.items()
            }


def run_source(context: 'Context', args: list, kwargs: dict, flags: list):
    """
    Take the input dict and impose global args/kwargs/flags with the following logic:
    - Use kwargs/flags as overriding keys in the input_context
    - Check the input dict if there is a key matching the arg and run that key
      - Additional arguments are assessed as
        - If the call is to a hook directly, then inject that as an argument
        - If the call is to a block of hooks then call the next hook key
    - Otherwise run normally (ie full parsing).

    An exception exists for if the last arg is `help` in which case that level's help
    is called and exited 0.
    """
    # Tackle is called both through the CLI and as a package and so to preserve args /
    # kwargs we
    if context.global_args is not None:
        args = args + context.global_args
        context.global_args = None

    # Global kwargs/flags are immediately consumed and injected into the kwargs/flags
    if context.global_kwargs is not None:
        kwargs.update(context.global_kwargs)
        context.global_kwargs = None

    if context.global_flags is not None:
        kwargs.update({i: True for i in context.global_flags})
        context.global_flags = None

    update_input_context_with_kwargs(context=context, kwargs=kwargs)

    for i in flags:
        # Process flags by setting key to true
        context.input_context.update({i: True})

    if len(args) >= 1:
        # TODO: Implement help
        # `help` which will always be the last arg
        # if args[-1] == 'help':
        #     # Calling help will exit 0. End of the line.
        #     run_help(context, context.input_context, args[:-1])

        # Loop through all args
        for i in args:
            # Remove any arrows on the first level keys
            first_level_compact_keys = [
                k[:-2] for k, _ in context.input_context.items() if k.endswith('->')
            ]
            if i in first_level_compact_keys:
                arg_key_value = context.input_context[i + '->']
                if isinstance(arg_key_value, str):
                    # We have a compact hook so nothing else to traverse
                    break

            elif i in context.input_context:
                context.key_path.append(i)
                walk_sync(context, context.input_context[i].copy())
                context.key_path.pop()

            elif i in context.provider_hooks:

                raise NotImplementedError
            else:
                raise ValueError(f"Argument {i} not found as key in input.")
        return

    if len(context.input_context) == 0:
        raise EmptyTackleFileException(
            # TODO improve -> Should give help by default?
            f"Only functions are declared in {context.input_string} tackle file. Must"
            f" provide an argument such as [] or run `tackle {context.input_string}"
            f" help` to see more options.",
            context=context,
        )
    else:
        walk_sync(context, context.input_context.copy())


def function_walk(
    self: Type[BaseFunction],
    input_element: Union[list, dict],
    return_: Union[list, dict] = None,
) -> Any:
    """
    Walk an input_element for a function and either return the whole context or one or
    many returnable string keys. Function is meant to be implanted into a function
    object and called either as `exec` or some other arbitrary method.
    """
    if self.public_context:
        existing_context = self.public_context.copy()
        existing_context.update(self.existing_context)
    else:
        existing_context = {}

    for i in self.function_fields:
        existing_context.update({i: getattr(self, i)})

    tmp_context = Context(
        provider_hooks=self.provider_hooks,
        existing_context=existing_context,
        public_context={},
        input_context=input_element,
        key_path=[],
        no_input=self.no_input,
        calling_directory=self.calling_directory,
        calling_file=self.calling_file,
    )
    walk_sync(context=tmp_context, element=input_element.copy())

    if return_:
        if isinstance(return_, str):
            if return_ in tmp_context.public_context:
                return tmp_context.public_context[return_]
            else:
                raise FunctionCallException(
                    f"Return value '{return_}' is not found " f"in output.",
                    function=self,
                )
        elif isinstance(return_, list):
            if isinstance(tmp_context, list):
                # TODO: This is not implemented (ie list outputs)
                raise FunctionCallException(
                    f"Can't have list return {return_} for " f"list output.",
                    function=self,
                )
            output = {}
            for i in return_:
                # Can only return top level keys right now
                if i in tmp_context.public_context:
                    output[i] = tmp_context.public_context[i]
                else:
                    raise FunctionCallException(
                        f"Return value '{i}' in return {return_} not found in output.",
                        function=self,
                    )
            return tmp_context.public_context[return_]
        else:
            raise NotImplementedError(f"Return must be of list or string {return_}.")
    return tmp_context.public_context


def create_function_model(
    context: 'Context', func_name: str, func_dict: dict
) -> Type[BaseFunction]:
    """Create a model from the function input dict."""
    if 'extends' in func_dict:
        base_hook = context.provider_hooks[func_dict['extends']]
        func_dict = {**base_hook, **func_dict}
        func_dict.pop('extends')

    # fmt: off
    # Validate raw input params against pydantic object where values will be used later
    function_input = FunctionInput(
        exec_=func_dict.pop('exec') if 'exec' in func_dict else None,
        return_=func_dict.pop('return') if 'return' in func_dict else None,
        args=func_dict.pop('args') if 'args' in func_dict else [],
        render_exclude=func_dict.pop(
            'render_exclude') if 'render_exclude' in func_dict else [],
        # validators_=func_dict.pop('validators') if 'validators' in func_dict else None,
        # methods_=func_dict.pop('methods') if 'methods' in func_dict else None,
    )
    # fmt: on

    new_func = {'hook_type': func_name[:-2], 'function_fields': []}
    # Create function fields from anything left over in the function dict
    for k, v in func_dict.items():
        if isinstance(v, dict):
            new_func[k] = Field(**v)
        elif isinstance(v, (str, int, float, bool)):
            new_func[k] = v
        elif isinstance(v, list):
            new_func[k] = (list, v)
        new_func['function_fields'].append(k)

    Function = create_model(
        func_name[:-2],
        __base__=BaseFunction,
        **new_func,
        **function_input.dict(include={'args', 'render_exclude'}),
    )
    setattr(
        Function,
        'exec',
        partialmethod(function_walk, function_input.exec_, function_input.return_),
    )

    context.env_.filters[func_name[:-2]] = Function(
        existing_context={},
        no_input=context.no_input,
    ).wrapped_exec

    return Function


def extract_functions(context: 'Context'):
    for k, v in context.input_context.copy().items():
        if re.match(r'^[a-zA-Z0-9\_]*(<\-|<\_)$', k):
            Function = create_function_model(context, k, v)
            context.provider_hooks[k[:-2]] = Function
            context.input_context.pop(k)


def extract_base_file(context: 'Context'):
    """Read the tackle file and initialize input_context."""
    if context.find_in_parent:
        path = find_in_parent(context.input_dir, context.input_file)
        context.input_file = os.path.basename(path)
        context.input_dir = os.path.dirname(path)
    else:
        path = os.path.join(context.input_dir, context.input_file)

    # Preserve the calling file which should be carried over from tackle calls
    if context.calling_file is None:
        context.calling_file = context.input_file
        # context.calling_file = path
    context.current_file = path

    context.input_context = read_config_file(path)
    if context.input_context is None:
        raise EmptyTackleFileException(
            f"No tackle file found at {path}.", context=context
        )

    if isinstance(context.input_context, list):
        # Change output to empty list
        context.public_context = []
        context.private_context = []
    else:
        extract_functions(context)
        context.public_context = {}
        context.private_context = {}

    # Check if there is a hooks directory in the provider being run and import the hooks
    input_dir_contents = os.listdir(context.input_dir)
    if 'hooks' in input_dir_contents:
        with work_in(context.input_dir):
            context.provider_hooks.import_from_path(context.input_dir)
        for i in context.provider_hooks.new_functions:
            try:
                context.env_.filters[i] = create_function_model(
                    context=context,
                    func_name=i,
                    func_dict=context.provider_hooks[i].dict(),
                )().wrapped_exec
            except KeyError:
                # TODO: This is odd - when running
                #  tackle/providers/tackle/tests/test_provider_tackle_import.py::test_provider_hook_import_func_provider_import
                #  There is an error from an adjacent test object that is thrown here
                #  only when all the tests are run. tackle-box/issues/61
                pass
        context.provider_hooks._new_functions = []

    # TODO: Experimental feature that could be integrated later
    # # Extract handlers
    # for k, v in list(input_context.items()):
    #     if k.startswith('__'):
    #         # Run pre-execution handlers and remove from input
    #         run_handler(context, k[2:], v)
    #         input_context.pop(k)
    #     if k.endswith('__'):
    #         # Store post-execution handlers and remove from input
    #         # TODO: Execute post exec handlers
    #         context.post_exec_handlers.append({k[:-2]: v})
    #         input_context.pop(k)


def import_local_provider_source(context: 'Context', provider_dir: str):
    """
    Import a provider from a path by checking if the provider has a tackle file and
    returning a path.
    """
    context.input_dir = provider_dir
    if context.input_file is None:
        context.input_file = find_tackle_file(provider_dir)

    if context.directory:
        context.input_file = os.path.join(context.input_file, context.directory)

    extract_base_file(context)


def update_source(context: 'Context'):
    """
    Locate the repository directory from a template reference. This is the main parser
    for determining the source of the context and calls the succeeding parsing
    functions. The parsing order has the following order of precedence.

    If the template wasn't given then use the file in that parent directory.
    If the template refers to a zip file or zip url, download / unzip as the context.
    If the template refers to a repository URL, clone it.
    If the template refers to a file, use that as the context.
    """
    args, kwargs, flags = unpack_args_kwargs_string(context.input_string)

    first_arg = args[0]
    # Remove first args it will be consumed and no longer relevant
    args.pop(0)

    # Zipfile
    if first_arg.lower().endswith('.zip'):
        unzipped_dir = unzip(
            zip_uri=first_arg,
            clone_to_dir=settings.tackle_dir,
            no_input=context.no_input,
            password=context.password,  # TODO: RM - Should prompt?
        )
        import_local_provider_source(context, unzipped_dir)
    # Repo
    elif is_repo_url(first_arg):
        provider_dir = get_repo_source(first_arg, context.checkout)
        import_local_provider_source(context, provider_dir)
    # Directory
    elif is_directory_with_tackle(first_arg):
        # Special case where the input is a path to a directory. Need to override some
        # settings that would normally get populated by zip / repo refs. Does not need
        # a file reference as otherwise would be given absolute path to tackle file.
        context.input_file = os.path.basename(find_tackle_file(first_arg))
        context.input_dir = Path(first_arg).absolute()

        # Load the base file into input_context
        extract_base_file(context)
    # File
    elif is_file(first_arg):
        context.input_file = os.path.basename(first_arg)
        context.input_dir = Path(first_arg).parent.absolute()
        extract_base_file(context)
    # Search in parent
    else:
        # Lastly we check if there is a key in the parent that matches the arg. This is
        # basically the fallback logic but need to raise error right away if the key
        # does not exist so we don't have to catch it with context later.
        tackle_file = find_nearest_tackle_file()
        if tackle_file is None:
            raise UnknownSourceException(
                f"Could not find source = {first_arg}", context=context
            )

        context.input_file = os.path.basename(tackle_file)
        context.input_dir = Path(tackle_file).parent.absolute()
        extract_base_file(context)

        if first_arg not in context.input_context:
            context.input_file = first_arg
            raise UnknownSourceException(
                f"Could not find source = {first_arg} or as "
                f"key in parent tackle file.",
                context=context,
            )
        args.insert(0, first_arg)

    # We always change directory into the source that is being called. Needs to be this
    # or would be very confusing if writing a provider to always refer to it's own path.
    with work_in(context.input_dir):
        # Main parsing logic
        run_source(context, args, kwargs, flags)



"""Tests `import` in the `tackle.providers.tackle.hooks.import` hook."""
import os

import pytest

from tackle import tackle
from tackle.models import Context
from tackle.parser import update_source

FIXTURES = [
    # # Handlers are experimental features
    # 'handler-str.yaml',
    # 'handler-list-compact.yaml',
    # 'expanded-string.yaml',
    # 'expanded-list-dict.yaml',
    'local.yaml',
    'function-import.yaml',
]


@pytest.mark.parametrize("target", FIXTURES)
def test_provider_system_hook_import(change_dir, target):
    """Run the source and check that the hooks imported the demo module."""
    context = Context(input_string=target)
    num_providers = len(context.provider_hooks.keys())
    update_source(context)
    assert num_providers < len(context.provider_hooks.keys())
    # assert 'tackle.providers.tackle-demos' in context.providers.hook_types


def test_provider_system_hook_import_local(change_dir):
    """Assert local import of hook is valid."""
    o = tackle('local.yaml')
    assert o['stuff'] == 'thing'


def test_provider_hook_import_func_provider_import(change_dir):
    """Assert local import of hook is valid."""
    os.chdir('func-provider')
    o = tackle()
    assert o['compact'] == 'a-default'
    assert o['jinja_extension_default'] == 'a-default'
    assert o['jinja_extension'] == 'things'
    assert o['jinja_filter'] == 'things'



import os
import sys
import re
import inspect
import threading
import subprocess
import importlib.machinery
from abc import ABC

from pydantic import (
    BaseModel,
    SecretStr,
    Field,
    Extra,
    validator,
    validate_model,
    ConfigError,
)
from pydantic.main import ModelMetaclass
from pydantic.fields import UndefinedType, ModelField
from jinja2 import Environment, StrictUndefined
from jinja2.ext import Extension
from typing import Any, Union, Optional, Dict, Type, Tuple
import logging
import random
import string

from tackle.utils.paths import listdir_absolute
from tackle.utils.dicts import get_readable_key_path
from tackle.render import wrap_jinja_braces
from tackle.utils.files import read_config_file

logger = logging.getLogger(__name__)


class LazyImportHook(BaseModel):
    """Object to hold hook metadata so that it can be imported only when called."""

    hooks_path: str
    mod_name: str
    provider_name: str
    hook_type: str

    def wrapped_exec(self, **kwargs):
        kwargs['provider_hooks'].import_with_fallback_install(
            mod_name=self.mod_name,
            path=self.hooks_path,
        )
        return kwargs['provider_hooks'][self.hook_type].wrapped_exec()


class ProviderHooks(dict):
    """Dict with hook_types as keys mapped to their corresponding objects."""

    # List to keep track of new functions which need to be updated into the jinja env's
    #  filters so that a hook can be called that way.
    new_functions: list = []

    def __init__(self, *args, **kwargs):
        super(ProviderHooks, self).__init__(*args, **kwargs)
        # https://github.com/robcxyz/tackle-box/issues/43
        # run_id used to make the namespace unique when running multiple runs of tackle
        # as in running batches of tests to prevent duplicate validator import errors
        self._run_id = ''.join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(4)
        )
        self.import_native_providers()

    def import_hook_from_path(
        self,
        mod_name: str,
        file_path: str,
    ):
        """Import a single hook from a path."""
        file_split = os.path.basename(file_path).split('.')
        file_base = file_split[0]
        file_extension = file_split[-1]
        # Maintaining cookiecutter support here as it might have a `hooks` dir.
        if file_base in ('pre_gen_project', 'post_gen_project', '__pycache__'):
            return
        if file_extension == 'pyc':
            return
        if file_extension in ('yaml', 'yml'):
            file_contents = read_config_file(file_path)
            for k, v in file_contents.items():
                if re.match(r'^[a-zA-Z0-9\_]*(<\-|<\_)$', k):
                    self[k[:-2]] = LazyBaseFunction(**v)
                    self.new_functions.append(k[:-2])
            return

        if os.path.basename(file_path).split('.')[-1] != "py":
            # Only import python files
            return

        # Use a unique RUN_ID to prevent duplicate validator errors
        # https://github.com/robcxyz/tackle-box/issues/43
        loader = importlib.machinery.SourceFileLoader(
            mod_name + '.hooks.' + file_base[0] + self._run_id, file_path
        )

        module = loader.load_module()

        for k, v in module.__dict__.items():
            if (
                isinstance(v, ModelMetaclass)
                and 'hook_type' in v.__fields__
                and k != 'BaseHook'
                and v.__fields__['hook_type'].default is not None
            ):
                self[v.__fields__['hook_type'].default] = v

    def import_hooks_from_dir(
        self,
        mod_name: str,
        path: str,
        skip_on_error: bool = False,
    ):
        """
        Import hooks from a directory. This is meant to be used by generically pointing to
         a hooks directory and importing all the relevant hooks into the context.
        """
        potential_hooks = listdir_absolute(path)
        for f in potential_hooks:
            if skip_on_error:
                try:
                    self.import_hook_from_path(mod_name, f)
                except (ModuleNotFoundError, ConfigError):
                    logger.debug(f"Skipping importing {f}")
                    continue
            else:
                self.import_hook_from_path(mod_name, f)

    def import_with_fallback_install(
        self, mod_name: str, path: str, skip_on_error: bool = False
    ):
        """
        Import a module and on import error, fallback on requirements file and try to
         import again.
        """
        try:
            self.import_hooks_from_dir(mod_name, path, skip_on_error)
        except ModuleNotFoundError:
            requirements_path = os.path.join(path, '..', 'requirements.txt')
            if os.path.isfile(requirements_path):
                # It is a convention of providers to have a requirements file at the base.
                # Install the contents if there was an import error
                subprocess.check_call(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        "--quiet",
                        "--disable-pip-version-check",
                        "-r",
                        requirements_path,
                    ]
                )
            self.import_hooks_from_dir(mod_name, path)

    def import_from_path(self, provider_path: str):
        """Append a provider with a given path."""
        provider_name = os.path.basename(provider_path)
        mod_name = 'tackle.providers.' + provider_name
        hooks_init_path = os.path.join(provider_path, 'hooks', '__init__.py')
        hooks_path = os.path.join(provider_path, 'hooks')

        # If the provider has an __init__.py in the hooks directory, import that
        # to check if there are any hook types declared there.  If there are, store
        # those references so that if the hook type is later called, the hook can
        # then be imported.
        if os.path.isfile(hooks_init_path):
            loader = importlib.machinery.SourceFileLoader(mod_name, hooks_init_path)
            mod = loader.load_module()
            hook_types = getattr(mod, 'hook_types', [])
            for h in hook_types:
                hook = LazyImportHook(
                    hooks_path=hooks_path,
                    mod_name=mod_name,
                    provider_name=provider_name,
                    hook_type=h,
                )
                self[h] = hook

        # This pass will import all the modules and extract hooks
        self.import_with_fallback_install(mod_name, hooks_path, skip_on_error=True)

    @staticmethod
    def get_native_provider_paths():
        """Get a list of paths to the native providers."""
        providers_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'providers'
        )
        native_providers = [
            os.path.join(providers_path, f)
            for f in os.listdir(providers_path)
            if os.path.isdir(os.path.join(providers_path, f)) and f != '__pycache__'
        ]
        return native_providers

    def import_native_providers(self):
        """Iterate through paths and import them."""
        native_provider_paths = self.get_native_provider_paths()
        for i in native_provider_paths:
            self.import_from_path(i)


class PartialModelMetaclass(ModelMetaclass):
    """
    Metaclass to allow partial model initialization.
    See https://github.com/samuelcolvin/pydantic/issues/1223#issuecomment-998160737
    """

    def __new__(
        meta: Type["PartialModelMetaclass"], *args: Any, **kwargs: Any
    ) -> "PartialModelMetaclass":
        try:
            cls = super(PartialModelMetaclass, meta).__new__(meta, *args, **kwargs)
        except Exception as e:
            raise e
        cls_init = cls.__init__
        # Because the class will be modified temporarily, need to lock __init__
        init_lock = threading.Lock()
        # To preserve identical hashes of temporary nested partial models,
        # only one instance of each temporary partial class can exist
        temporary_partial_classes: Dict[str, ModelMetaclass] = {}

        def __init__(self: BaseModel, *args: Any, **kwargs: Any) -> None:
            with init_lock:
                fields = self.__class__.__fields__
                fields_map: Dict[ModelField, Tuple[Any, bool]] = {}

                def optionalize(
                    fields: Dict[str, ModelField], *, restore: bool = False
                ) -> None:
                    for _, field in fields.items():
                        if not restore:
                            assert not isinstance(field.required, UndefinedType)
                            fields_map[field] = (field.type_, field.required)
                            field.required = False
                            if (
                                inspect.isclass(field.type_)
                                and issubclass(field.type_, BaseModel)
                                and not field.type_.__name__.startswith(
                                    "TemporaryPartial"
                                )
                            ):
                                # Assign a temporary type to optionalize to avoid
                                # modifying *other* classes
                                class_name = f"TemporaryPartial{field.type_.__name__}"
                                if class_name in temporary_partial_classes:
                                    field.type_ = temporary_partial_classes[class_name]
                                else:
                                    field.type_ = ModelMetaclass(
                                        class_name,
                                        (field.type_,),
                                        {},
                                    )
                                    temporary_partial_classes[class_name] = field.type_
                                optionalize(field.type_.__fields__)
                                # After replacing the field type, regenerate validators
                                field.populate_validators()
                        else:
                            # No need to recursively de-optionalize once original types
                            # are restored
                            field.type_, field.required = fields_map[field]

                # Make fields and fields of nested model types optional
                optionalize(fields)
                # Transform kwargs that are PartialModels to their dict() forms. This
                # will exclude `None` (see below) from the dictionary used to construct
                # the temporarily-partial model field, avoiding ValidationErrors of
                # type type_error.none.not_allowed.
                for kwarg, value in kwargs.items():
                    if value.__class__.__class__ is PartialModelMetaclass:
                        kwargs[kwarg] = value.dict()
                # Validation is performed in __init__, for which all fields are now optional
                if args:
                    for i, v in enumerate(args):
                        kwargs[self.args[i]] = v
                # cls is always initialized with kwargs not args.
                try:
                    cls_init(self, **kwargs)
                except Exception as e:
                    raise e
                # Restore requiredness
                optionalize(fields, restore=True)

        setattr(cls, "__init__", __init__)

        # Exclude unset (`None`) from dict(), which isn't allowed in the schema
        # but will be the default for non-required fields. This enables
        # PartialModel(**PartialModel().dict()) to work correctly.
        cls_dict = cls.dict

        def dict_exclude_unset(
            self: BaseModel, *args: Any, exclude_unset: bool = None, **kwargs: Any
        ) -> Dict[str, Any]:
            return cls_dict(self, *args, **kwargs, exclude_unset=True)

        cls.dict = dict_exclude_unset

        return cls


class StrictEnvironment(Environment):
    """Create strict Jinja2 environment.

    Jinja2 environment will raise error on undefined variable in template-
    rendering context.
    """

    def __init__(self, provider_hooks: dict, **kwargs):
        super(StrictEnvironment, self).__init__(undefined=StrictUndefined, **kwargs)
        # Import filters into environment
        for k, v in provider_hooks.items():
            if isinstance(v, LazyImportHook):
                self.filters[k] = v.wrapped_exec
            elif isinstance(v, LazyBaseFunction):
                self.filters[k] = v.wrapped_exec
            else:
                # Filters don't receive any context (public_context, etc)
                self.filters[k] = v().wrapped_exec


class BaseContext(BaseModel):
    """Shared items that are passed between hook calls and the main context."""

    no_input: bool = False
    verbose: bool = False

    input_context: Union[dict, list] = None
    public_context: Union[dict, list] = None
    private_context: Union[dict, list] = None
    temporary_context: Union[dict, list] = None
    existing_context: dict = {}

    key_path: list = []
    key_path_block: list = []
    provider_hooks: ProviderHooks = None

    calling_directory: str = None
    calling_file: str = None
    current_file: str = None

    env_: Any = None


class Context(BaseContext):
    """The main object that is being modified by parsing."""

    # Source
    password: SecretStr = None
    directory: str = None
    find_in_parent: bool = False
    input_string: str = None
    input_dir: str = None
    input_file: str = None

    # TODO: Change to version?
    checkout: str = Field(
        None,
        description="For inputs referencing remote repos, refers to a branch or tag.",
    )

    global_args: list = None
    global_kwargs: dict = None
    global_flags: list = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        # Allows for passing the providers between tackle runtimes
        if self.provider_hooks is None:
            self.provider_hooks = ProviderHooks()

        if self.env_ is None:
            self.env_ = StrictEnvironment(self.provider_hooks)

        if self.calling_directory is None:
            # Can be carried over from another context. Should only be initialized when
            # called from CLI.
            self.calling_directory = os.path.abspath(os.path.curdir)


class BaseHook(BaseContext, Extension, metaclass=PartialModelMetaclass):
    """
    Base hook class from which all other hooks inherit from to be discovered. There are
    a number of reserved keys that are used for logic such as `if` and `for` that are
    aliased to `if_` and `for_` to not collide with python reserved key words.
    """

    hook_type: str = Field(..., description="Name of the hook.")

    if_: Union[str, bool] = Field(None, render_by_default=True)
    when: Union[str, bool] = Field(None, render_by_default=True)
    else_: Any = Field(None)
    for_: Union[str, list] = Field(None, render_by_default=True)
    reverse: Union[str, bool] = Field(None, render_by_default=True)
    try_: Union[str, bool] = Field(None, render_by_default=True)
    except_: Any = Field(None, render_by_default=True)
    chdir: Optional[str] = Field(None, description="Name of the hook.")
    merge: Union[bool, str] = None
    confirm: Optional[Any] = None

    # Placeholder until help can be fully worked out
    help: str = None

    # Flag for whether being called directly (True) or by a jinja extension (False)
    is_hook_call: bool = False
    skip_import: bool = False
    hooks_path: str = None

    env_: Any = None

    args: list = []

    skip_output: bool = False
    # Fields that should not be rendered by default
    _render_exclude_default: set = {
        'input_context',
        'public_context',
        'hook_type',
        'else',
    }
    _render_exclude: set = {}
    _render_by_default: list = []

    # Used when rendering docs
    _doc_tags: list = []
    # For linking issues in the docs so others can potentially contribute
    _issue_numbers: list = []
    # Additional callout sections to be included at the top of the docs
    _notes: list = []
    # Allow hooks to be sorted in the docs
    _docs_order: int = 10  # Arbitrary high number so hooks can be sorted high or low
    # Parameterized return type description for docs
    _return_description: str = None

    @validator('if_', 'reverse', 'for_', 'merge')
    def wrap_bool_if_string(cls, v):
        return wrap_jinja_braces(v)

    # Per https://github.com/samuelcolvin/pydantic/issues/1577
    def __setattr__(self, key, val):
        """Override method to alias input fields."""
        if key in self.__config__.alias_to_fields:
            key = self.__config__.alias_to_fields[key]
        super().__setattr__(key, val)

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid
        validate_assignment = True
        fields = {
            'if_': 'if',
            'else_': 'else',
            'for_': 'for',
            'try_': 'try',
            'except_': 'except',
        }
        # Per https://github.com/samuelcolvin/pydantic/issues/1577
        # This is an issue until pydantic 1.9 is released and items can be set with
        # properties which will override the internal __setattr__ method that
        # disregards aliased fields
        alias_to_fields = {v: k for k, v in fields.items()}

    def exec(self) -> Any:
        raise NotImplementedError("Every hook needs an exec method.")

    # https://github.com/samuelcolvin/pydantic/issues/1864#issuecomment-679044432
    def validate(self, **kwargs):
        """Manual validation with mapping back aliases as they will be remapped."""
        # For some reason changes here apply to all instances of a class and so if a
        # hook is reused, it will fail because the special aliased fields are already
        # remapped. This is a hack.
        if 'if_' in self.__dict__:
            for k, v in self.Config.fields.items():
                self.__dict__[v] = self.__dict__[k]
                self.__dict__.pop(k)
        *_, validation_error = validate_model(self.__class__, self.__dict__)
        if validation_error:
            raise validation_error

    def wrapped_exec(self, *args, **kwargs):
        # Map args / kwargs
        if args:
            num_args = len(args)
            for i, v in enumerate(self.args):
                if i >= num_args:
                    break
                setattr(self, v, args[i])
            if len(args) > len(self.args):
                raise Exception(
                    f"Too many arguments in {get_readable_key_path(self.key_path)}"
                )
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.validate()
        return self.exec()


class FunctionInput(BaseModel):
    """Function input model."""

    exec_: Any = Field(None)
    return_: Union[str, list] = Field(None)

    args: list = None
    render_exclude: list = None
    validators: dict = None
    methods: dict = None
    public: bool = None
    extends: str = None

    class Config:
        extra = 'ignore'


class BaseFunction(BaseHook, FunctionInput, ABC):
    """Function input model."""

    function_fields: list


class LazyBaseFunction(BaseFunction, Extension, metaclass=PartialModelMetaclass):
    """
    Base function that declarative hooks are derived from and either imported when a
     tackle file is read (by searching in adjacent hooks directory) on init in local
     providers. Used by jinja extensions and filters.
    """

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow
        # TODO: Figure this out as it is needed in the combined jinja hook wrapped exec
        #  logic... In this case we already have it serialized?
        # Not sure what is going on here
        fields = {
            'if_': 'if_',
            'else_': 'else_',
            'for_': 'for_',
            'try_': 'try_',
            'except_': 'except_',
        }
